@@ -1,38 +1,23 @@
-#
 # Copyright (C) 2006-2008, 2013-2014 Red Hat, Inc.
 # Copyright (C) 2006 Daniel P. Berrange <berrange@redhat.com>
 #
-# This program is free software; you can redistribute it and/or modify
-# it under the terms of the GNU General Public License as published by
-# the Free Software Foundation; either version 2 of the License, or
-# (at your option) any later version.
-#
-# This program is distributed in the hope that it will be useful,
-# but WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-# GNU General Public License for more details.
-#
-# You should have received a copy of the GNU General Public License
-# along with this program; if not, write to the Free Software
-# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
-# MA 02110-1301 USA.
-#
-
-import logging
+# This work is licensed under the GNU GPLv2 or later.
+# See the COPYING file in the top-level directory.
 
 from gi.repository import GObject
 from gi.repository import Gtk
 from gi.repository import Gdk
 from gi.repository import GdkPixbuf
 
-import libvirt
-
-from virtinst import util
+from virtinst import log
+from virtinst import xmlutil
 
 from . import vmmenu
-from . import uiutil
+from .lib import uiutil
 from .baseclass import vmmGObjectUI
-from .graphwidgets import CellRendererSparkline
+from .connmanager import vmmConnectionManager
+from .engine import vmmEngine
+from .lib.graphwidgets import CellRendererSparkline
 
 # Number of data points for performance graphs
 GRAPH_LEN = 40
@@ -66,6 +51,10 @@ def _style_get_prop(widget, propname):
     return value.get_int()
 
 
+def _cmp(a, b):
+    return ((a > b) - (a < b))
+
+
 def _get_inspection_icon_pixbuf(vm, w, h):
     # libguestfs gives us the PNG data as a string.
     png_data = vm.inspection.icon
@@ -78,43 +67,28 @@ def _get_inspection_icon_pixbuf(vm, w, h):
         pb.write(png_data)
         pb.close()
         return pb.get_pixbuf()
-    except:
-        logging.exception("Error loading inspection icon data")
+    except Exception:  # pragma: no cover
+        log.exception("Error loading inspection icon data")
         vm.inspection.icon = None
         return None
 
 
 class vmmManager(vmmGObjectUI):
-    __gsignals__ = {
-        "action-show-connect": (GObject.SignalFlags.RUN_FIRST, None, []),
-        "action-show-domain": (GObject.SignalFlags.RUN_FIRST, None, [str, str]),
-        "action-show-about": (GObject.SignalFlags.RUN_FIRST, None, []),
-        "action-show-host": (GObject.SignalFlags.RUN_FIRST, None, [str]),
-        "action-show-preferences": (GObject.SignalFlags.RUN_FIRST, None, []),
-        "action-show-create": (GObject.SignalFlags.RUN_FIRST, None, [str]),
-        "action-suspend-domain": (GObject.SignalFlags.RUN_FIRST, None, [str, str]),
-        "action-resume-domain": (GObject.SignalFlags.RUN_FIRST, None, [str, str]),
-        "action-run-domain": (GObject.SignalFlags.RUN_FIRST, None, [str, str]),
-        "action-shutdown-domain": (GObject.SignalFlags.RUN_FIRST, None, [str, str]),
-        "action-reset-domain": (GObject.SignalFlags.RUN_FIRST, None, [str, str]),
-        "action-reboot-domain": (GObject.SignalFlags.RUN_FIRST, None, [str, str]),
-        "action-destroy-domain": (GObject.SignalFlags.RUN_FIRST, None, [str, str]),
-        "action-save-domain": (GObject.SignalFlags.RUN_FIRST, None, [str, str]),
-        "action-migrate-domain": (GObject.SignalFlags.RUN_FIRST, None, [str, str]),
-        "action-delete-domain": (GObject.SignalFlags.RUN_FIRST, None, [str, str]),
-        "action-clone-domain": (GObject.SignalFlags.RUN_FIRST, None, [str, str]),
-        "action-exit-app": (GObject.SignalFlags.RUN_FIRST, None, []),
-        "manager-closed": (GObject.SignalFlags.RUN_FIRST, None, []),
-        "manager-opened": (GObject.SignalFlags.RUN_FIRST, None, []),
-        "remove-conn": (GObject.SignalFlags.RUN_FIRST, None, [str]),
-    }
+    @classmethod
+    def get_instance(cls, parentobj):
+        try:
+            if not cls._instance:
+                cls._instance = vmmManager()
+            return cls._instance
+        except Exception as e:  # pragma: no cover
+            if not parentobj:
+                raise
+            parentobj.err.show_err(
+                    _("Error launching manager: %s") % str(e))
 
     def __init__(self):
         vmmGObjectUI.__init__(self, "manager.ui", "vmm-manager")
-
-        # Mapping of rowkey -> tree model rows to
-        # allow O(1) access instead of O(n)
-        self.rows = {}
+        self._cleanup_on_app_close()
 
         w, h = self.config.get_manager_window_size()
         self.topwin.set_default_size(w or 550, h or 550)
@@ -122,24 +96,26 @@ class vmmManager(vmmGObjectUI):
         self._window_size = None
 
         self.vmmenu = vmmenu.VMActionMenu(self, self.current_vm)
+        self.shutdownmenu = vmmenu.VMShutdownMenu(self, self.current_vm)
         self.connmenu = Gtk.Menu()
+        self.connmenu.get_accessible().set_name("conn-menu")
         self.connmenu_items = {}
 
         self.builder.connect_signals({
             "on_menu_view_guest_cpu_usage_activate":
-                    self.toggle_stats_visible_guest_cpu,
+            self.toggle_stats_visible_guest_cpu,
             "on_menu_view_host_cpu_usage_activate":
-                    self.toggle_stats_visible_host_cpu,
+            self.toggle_stats_visible_host_cpu,
             "on_menu_view_memory_usage_activate":
-                    self.toggle_stats_visible_memory_usage,
-            "on_menu_view_disk_io_activate" :
-                    self.toggle_stats_visible_disk,
+            self.toggle_stats_visible_memory_usage,
+            "on_menu_view_disk_io_activate":
+            self.toggle_stats_visible_disk,
             "on_menu_view_network_traffic_activate":
-                    self.toggle_stats_visible_network,
+            self.toggle_stats_visible_network,
 
             "on_vm_manager_delete_event": self.close,
             "on_vmm_manager_configure_event": self.window_resized,
-            "on_menu_file_add_connection_activate": self.new_conn,
+            "on_menu_file_add_connection_activate": self.open_newconn,
             "on_menu_new_vm_activate": self.new_vm,
             "on_menu_file_quit_activate": self.exit_app,
             "on_menu_file_close_activate": self.close,
@@ -153,7 +129,7 @@ class vmmManager(vmmGObjectUI):
             "on_menu_edit_delete_activate": self.do_delete,
             "on_menu_host_details_activate": self.show_host,
 
-            "on_vm_list_row_activated": self.show_vm,
+            "on_vm_list_row_activated": self.row_activated,
             "on_vm_list_button_press_event": self.popup_vm_menu_button,
             "on_vm_list_key_press_event": self.popup_vm_menu_key,
 
@@ -184,10 +160,16 @@ class vmmManager(vmmGObjectUI):
 
         # Initialize stat polling columns based on global polling
         # preferences (we want signal handlers for this)
-        self.enable_polling(COL_GUEST_CPU)
-        self.enable_polling(COL_DISK)
-        self.enable_polling(COL_NETWORK)
-        self.enable_polling(COL_MEM)
+        self._config_polling_change_cb(COL_GUEST_CPU)
+        self._config_polling_change_cb(COL_DISK)
+        self._config_polling_change_cb(COL_NETWORK)
+        self._config_polling_change_cb(COL_MEM)
+
+        connmanager = vmmConnectionManager.get_instance()
+        connmanager.connect("conn-added", self._conn_added)
+        connmanager.connect("conn-removed", self._conn_removed)
+        for conn in connmanager.conns.values():
+            self._conn_added(connmanager, conn)
 
 
     ##################
@@ -200,34 +182,34 @@ class vmmManager(vmmGObjectUI):
         if vis:
             return
 
-        logging.debug("Showing manager")
+        log.debug("Showing manager")
         if self.prev_position:
             self.topwin.move(*self.prev_position)
             self.prev_position = None
 
-        self.emit("manager-opened")
+        vmmEngine.get_instance().increment_window_counter()
 
     def close(self, src_ignore=None, src2_ignore=None):
         if not self.is_visible():
             return
 
-        logging.debug("Closing manager")
+        log.debug("Closing manager")
         self.prev_position = self.topwin.get_position()
         self.topwin.hide()
-        self.emit("manager-closed")
+        vmmEngine.get_instance().decrement_window_counter()
 
         return 1
 
 
     def _cleanup(self):
-        self.rows = None
-
         self.diskcol = None
         self.guestcpucol = None
         self.memcol = None
         self.hostcpucol = None
         self.netcol = None
 
+        self.shutdownmenu.destroy()
+        self.shutdownmenu = None
         self.vmmenu.destroy()
         self.vmmenu = None
         self.connmenu.destroy()
@@ -237,14 +219,11 @@ class vmmManager(vmmGObjectUI):
         if self._window_size:
             self.config.set_manager_window_size(*self._window_size)
 
-
-    def is_visible(self):
-        return bool(self.topwin.get_visible())
-
     def set_startup_error(self, msg):
         self.widget("vm-notebook").set_current_page(1)
         self.widget("startup-error-label").set_text(msg)
 
+
     ################
     # Init methods #
     ################
@@ -270,16 +249,16 @@ class vmmManager(vmmGObjectUI):
         # that disable the associated vmlist widgets if reporting is disabled
         self.add_gsettings_handle(
             self.config.on_stats_enable_cpu_poll_changed(
-                self.enable_polling, COL_GUEST_CPU))
+                self._config_polling_change_cb, COL_GUEST_CPU))
         self.add_gsettings_handle(
             self.config.on_stats_enable_disk_poll_changed(
-                self.enable_polling, COL_DISK))
+                self._config_polling_change_cb, COL_DISK))
         self.add_gsettings_handle(
             self.config.on_stats_enable_net_poll_changed(
-                self.enable_polling, COL_NETWORK))
+                self._config_polling_change_cb, COL_NETWORK))
         self.add_gsettings_handle(
             self.config.on_stats_enable_memory_poll_changed(
-                self.enable_polling, COL_MEM))
+                self._config_polling_change_cb, COL_MEM))
 
         self.toggle_guest_cpu_usage_visible_widget()
         self.toggle_host_cpu_usage_visible_widget()
@@ -292,9 +271,8 @@ class vmmManager(vmmGObjectUI):
         self.widget("vm-new").set_icon_name("vm_new")
         self.widget("vm-open").set_icon_name("icon_console")
 
-        menu = vmmenu.VMShutdownMenu(self, self.current_vm)
         self.widget("vm-shutdown").set_icon_name("system-shutdown")
-        self.widget("vm-shutdown").set_menu(menu)
+        self.widget("vm-shutdown").set_menu(self.shutdownmenu)
 
         tool = self.widget("vm-toolbar")
         tool.set_property("icon-size", Gtk.IconSize.LARGE_TOOLBAR)
@@ -302,27 +280,26 @@ class vmmManager(vmmGObjectUI):
             c.set_homogeneous(False)
 
     def init_context_menus(self):
-        def add_to_menu(idx, text, icon, cb):
+        def add_to_menu(idx, text, cb):
             if text[0:3] == 'gtk':
                 item = Gtk.ImageMenuItem.new_from_stock(text, None)
             else:
                 item = Gtk.ImageMenuItem.new_with_mnemonic(text)
-            if icon:
-                item.set_image(icon)
             if cb:
                 item.connect("activate", cb)
+            item.get_accessible().set_name("conn-%s" % idx)
             self.connmenu.add(item)
             self.connmenu_items[idx] = item
 
         # Build connection context menu
-        add_to_menu("create", Gtk.STOCK_NEW, None, self.new_vm)
-        add_to_menu("connect", Gtk.STOCK_CONNECT, None, self.open_conn)
-        add_to_menu("disconnect", Gtk.STOCK_DISCONNECT, None,
+        add_to_menu("create", Gtk.STOCK_NEW, self.new_vm)
+        add_to_menu("connect", Gtk.STOCK_CONNECT, self.open_conn)
+        add_to_menu("disconnect", Gtk.STOCK_DISCONNECT,
                       self.close_conn)
         self.connmenu.add(Gtk.SeparatorMenuItem())
-        add_to_menu("delete", Gtk.STOCK_DELETE, None, self.do_delete)
+        add_to_menu("delete", Gtk.STOCK_DELETE, self.do_delete)
         self.connmenu.add(Gtk.SeparatorMenuItem())
-        add_to_menu("details", _("D_etails"), None, self.show_host)
+        add_to_menu("details", _("D_etails"), self.show_host)
         self.connmenu.show_all()
 
     def init_vmlist(self):
@@ -413,10 +390,15 @@ class vmmManager(vmmGObjectUI):
         model.set_sort_func(COL_NETWORK, self.vmlist_network_usage_sorter)
         model.set_sort_column_id(COL_NAME, Gtk.SortType.ASCENDING)
 
+
     ##################
     # Helper methods #
     ##################
 
+    @property
+    def model(self):
+        return self.widget("vm-list").get_model()
+
     def current_row(self):
         return uiutil.get_list_selected_row(self.widget("vm-list"))
 
@@ -431,68 +413,78 @@ class vmmManager(vmmGObjectUI):
         row = self.current_row()
         if not row:
             return None
-
         handle = row[ROW_HANDLE]
         if row[ROW_IS_CONN]:
             return handle
-        else:
-            return handle.conn
+        return handle.conn
 
-    def current_conn_uri(self, default_selection=False):
-        vmlist = self.widget("vm-list")
-        model = vmlist.get_model()
+    def get_row(self, conn_or_vm):
+        def _walk(model, rowiter, obj):
+            while rowiter:
+                row = model[rowiter]
+                if row[ROW_HANDLE] == obj:
+                    return row
+                if model.iter_has_child(rowiter):
+                    ret = _walk(model, model.iter_nth_child(rowiter, 0), obj)
+                    if ret:
+                        return ret
+                rowiter = model.iter_next(rowiter)
 
-        conn = self.current_conn()
-        if conn is None and default_selection:
-            # Nothing selected, use first connection row
-            for row in model:
-                if row[ROW_IS_CONN]:
-                    conn = row[ROW_HANDLE]
-                    break
+        if not len(self.model):
+            return None
+        return _walk(self.model, self.model.get_iter_first(), conn_or_vm)
 
-        if conn:
-            return conn.get_uri()
-        return None
 
     ####################
     # Action listeners #
     ####################
 
-    def window_resized(self, ignore, event):
+    def window_resized(self, ignore, ignore2):
         if not self.is_visible():
             return
-        self._window_size = (event.width, event.height)
+        self._window_size = self.topwin.get_size()
 
     def exit_app(self, src_ignore=None, src2_ignore=None):
-        self.emit("action-exit-app")
+        vmmEngine.get_instance().exit_app()
 
-    def new_conn(self, src_ignore=None):
-        self.emit("action-show-connect")
+    def open_newconn(self, _src):
+        from .createconn import vmmCreateConn
+        vmmCreateConn.get_instance(self).show(self.topwin)
 
-    def new_vm(self, src_ignore=None):
-        self.emit("action-show-create", self.current_conn_uri())
+    def new_vm(self, _src):
+        from .createvm import vmmCreateVM
+        conn = self.current_conn()
+        vmmCreateVM.show_instance(self, conn and conn.get_uri() or None)
 
-    def show_about(self, src_ignore):
-        self.emit("action-show-about")
+    def show_about(self, _src):
+        from .about import vmmAbout
+        vmmAbout.show_instance(self)
 
     def show_preferences(self, src_ignore):
-        self.emit("action-show-preferences")
+        from .preferences import vmmPreferences
+        vmmPreferences.show_instance(self)
 
-    def show_host(self, src_ignore):
-        uri = self.current_conn_uri(default_selection=True)
-        self.emit("action-show-host", uri)
+    def show_host(self, _src):
+        from .host import vmmHost
+        conn = self.current_conn()
+        vmmHost.show_instance(self, conn)
 
-    def show_vm(self, ignore, ignore2=None, ignore3=None):
+    def show_vm(self, _src):
+        vmmenu.VMActionUI.show(self, self.current_vm())
+
+    def row_activated(self, _src, *args):
+        ignore = args
         conn = self.current_conn()
         vm = self.current_vm()
         if conn is None:
-            return
+            return  # pragma: no cover
 
         if vm:
-            self.emit("action-show-domain", conn.get_uri(), vm.get_connkey())
+            self.show_vm(_src)
+        elif conn.is_disconnected():
+            self.open_conn()
         else:
-            if not self.open_conn():
-                self.emit("action-show-host", conn.get_uri())
+            self.show_host(_src)
 
     def do_delete(self, ignore=None):
         conn = self.current_conn()
@@ -500,18 +492,15 @@ class vmmManager(vmmGObjectUI):
         if vm is None:
             self._do_delete_conn(conn)
         else:
-            self.emit("action-delete-domain", conn.get_uri(), vm.get_connkey())
+            vmmenu.VMActionUI.delete(self, vm)
 
     def _do_delete_conn(self, conn):
-        if conn is None:
-            return
-
         result = self.err.yes_no(_("This will remove the connection:\n\n%s\n\n"
                                    "Are you sure?") % conn.get_uri())
         if not result:
             return
 
-        self.emit("remove-conn", conn.get_uri())
+        vmmConnectionManager.get_instance().remove_conn(conn.get_uri())
 
     def set_pause_state(self, state):
         src = self.widget("vm-pause")
@@ -529,34 +518,14 @@ class vmmManager(vmmGObjectUI):
         self.set_pause_state(not do_pause)
 
         if do_pause:
-            self.pause_vm(None)
+            vmmenu.VMActionUI.suspend(self, self.current_vm())
         else:
-            self.resume_vm(None)
+            vmmenu.VMActionUI.resume(self, self.current_vm())
 
     def start_vm(self, ignore):
-        vm = self.current_vm()
-        if vm is None:
-            return
-        self.emit("action-run-domain", vm.conn.get_uri(), vm.get_connkey())
-
-    def poweroff_vm(self, ignore):
-        vm = self.current_vm()
-        if vm is None:
-            return
-        self.emit("action-shutdown-domain",
-            vm.conn.get_uri(), vm.get_connkey())
-
-    def pause_vm(self, ignore):
-        vm = self.current_vm()
-        if vm is None:
-            return
-        self.emit("action-suspend-domain", vm.conn.get_uri(), vm.get_connkey())
-
-    def resume_vm(self, ignore):
-        vm = self.current_vm()
-        if vm is None:
-            return
-        self.emit("action-resume-domain", vm.conn.get_uri(), vm.get_connkey())
+        vmmenu.VMActionUI.run(self, self.current_vm())
+    def poweroff_vm(self, _src):
+        vmmenu.VMActionUI.shutdown(self, self.current_vm())
 
     def close_conn(self, ignore):
         conn = self.current_conn()
@@ -566,78 +535,66 @@ class vmmManager(vmmGObjectUI):
     def open_conn(self, ignore=None):
         conn = self.current_conn()
         if conn.is_disconnected():
+            conn.connect_once("open-completed", self._conn_open_completed_cb)
             conn.open()
             return True
 
+    def _conn_open_completed_cb(self, _conn, ConnectError):
+        if ConnectError:
+            msg, details, title = ConnectError
+            self.err.show_err(msg, details, title)
+
 
     ####################################
     # VM add/remove management methods #
     ####################################
 
-    def vm_row_key(self, vm):
-        return vm.get_uuid() + ":" + vm.conn.get_uri()
-
-    def vm_added(self, conn, connkey):
-        vm = conn.get_vm(connkey)
-        if not vm:
-            return
-
-        row_key = self.vm_row_key(vm)
-        if row_key in self.rows:
-            return
-
-        row = self._build_row(None, vm)
-        parent = self.rows[conn.get_uri()].iter
-        model = self.widget("vm-list").get_model()
-        _iter = model.append(parent, row)
-        path = model.get_path(_iter)
-        self.rows[row_key] = model[path]
+    def vm_added(self, conn, vm):
+        vm_row = self._build_row(None, vm)
+        conn_row = self.get_row(conn)
+        self.model.append(conn_row.iter, vm_row)
 
         vm.connect("state-changed", self.vm_changed)
         vm.connect("resources-sampled", self.vm_row_updated)
         vm.connect("inspection-changed", self.vm_inspection_changed)
 
         # Expand a connection when adding a vm to it
-        self.widget("vm-list").expand_row(model.get_path(parent), False)
+        self.widget("vm-list").expand_row(conn_row.path, False)
 
-    def vm_removed(self, conn, connkey):
-        vmlist = self.widget("vm-list")
-        model = vmlist.get_model()
-
-        parent = self.rows[conn.get_uri()].iter
-        for row in range(model.iter_n_children(parent)):
-            vm = model[model.iter_nth_child(parent, row)][ROW_HANDLE]
-            if vm.get_connkey() == connkey:
-                model.remove(model.iter_nth_child(parent, row))
-                del self.rows[self.vm_row_key(vm)]
+    def vm_removed(self, conn, vm):
+        parent = self.get_row(conn).iter
+        for rowidx in range(self.model.iter_n_children(parent)):
+            rowiter = self.model.iter_nth_child(parent, rowidx)
+            if self.model[rowiter][ROW_HANDLE] == vm:
+                self.model.remove(rowiter)
                 break
 
     def _build_conn_hint(self, conn):
         hint = conn.get_uri()
         if conn.is_disconnected():
-            hint += " (%s)" % _("Double click to connect")
+            hint = _("%(uri)s (Double click to connect)") % {"uri": conn.get_uri()}
         return hint
 
     def _build_conn_markup(self, conn, name):
-        name = util.xml_escape(name)
+        name = xmlutil.xml_escape(name)
         text = name
         if conn.is_disconnected():
-            text += " - " + _("Not Connected")
+            text = _("%(connection)s - Not Connected") % {"connection": name}
         elif conn.is_connecting():
-            text += " - " + _("Connecting...")
+            text = _("%(connection)s - Connecting...") % {"connection": name}
 
         markup = "<span size='smaller'>%s</span>" % text
         return markup
 
     def _build_conn_color(self, conn):
-        color = "#000000"
+        color = None
         if conn.is_disconnected():
-            color = "#5b5b5b"
+            color = self.config.color_insensitive
         return color
 
     def _build_vm_markup(self, name, status):
         domtext     = ("<span size='smaller' weight='bold'>%s</span>" %
-                       util.xml_escape(name))
+                       xmlutil.xml_escape(name))
         statetext   = "<span size='smaller'>%s</span>" % status
         return domtext + "\n" + statetext
 
@@ -665,7 +622,7 @@ class vmmManager(vmmGObjectUI):
         row.insert(ROW_SORT_KEY, name)
         row.insert(ROW_MARKUP, markup)
         row.insert(ROW_STATUS_ICON, status_icon)
-        row.insert(ROW_HINT, util.xml_escape(hint))
+        row.insert(ROW_HINT, xmlutil.xml_escape(hint))
         row.insert(ROW_IS_CONN, bool(conn))
         row.insert(ROW_IS_CONN_CONNECTED,
                    bool(conn) and not conn.is_disconnected())
@@ -676,43 +633,42 @@ class vmmManager(vmmGObjectUI):
 
         return row
 
-    def add_conn(self, engine_ignore, conn):
-        # Called from engine.py signal conn-added
-
+    def _conn_added(self, _src, conn):
         # Make sure error page isn't showing
         self.widget("vm-notebook").set_current_page(0)
+        if self.get_row(conn):
+            return  # pragma: no cover
 
-        if conn.get_uri() in self.rows:
-            return
-
-        model = self.widget("vm-list").get_model()
-        row = self._build_row(conn, None)
-        _iter = model.append(None, row)
-        path = model.get_path(_iter)
-        self.rows[conn.get_uri()] = model[path]
+        conn_row = self._build_row(conn, None)
+        self.model.append(None, conn_row)
 
         conn.connect("vm-added", self.vm_added)
         conn.connect("vm-removed", self.vm_removed)
         conn.connect("resources-sampled", self.conn_row_updated)
         conn.connect("state-changed", self.conn_state_changed)
 
-    def remove_conn(self, engine_ignore, uri):
-        # Called from engine.py signal conn-removed
+        for vm in conn.list_vms():
+            self.vm_added(conn, vm)
 
-        model = self.widget("vm-list").get_model()
-        parent = self.rows[uri].iter
+    def _remove_child_rows(self, row):
+        child = self.model.iter_children(row.iter)
+        while child is not None:  # pragma: no cover
+            # vm-removed signals should handle this, this is a fallback
+            # in case something goes wrong
+            self.model.remove(child)
+            child = self.model.iter_children(row.iter)
 
-        if parent is None:
+    def _conn_removed(self, _src, uri):
+        conn_row = None
+        for row in self.model:
+            if row[ROW_IS_CONN] and row[ROW_HANDLE].get_uri() == uri:
+                conn_row = row
+                break
+        if conn_row is None:  # pragma: no cover
             return
 
-        child = model.iter_children(parent)
-        while child is not None:
-            del self.rows[self.vm_row_key(model[child][ROW_HANDLE])]
-            model.remove(child)
-            child = model.iter_children(parent)
-        model.remove(parent)
-
-        del self.rows[uri]
+        self._remove_child_rows(conn_row)
+        self.model.remove(conn_row.iter)
 
 
     #############################
@@ -720,15 +676,15 @@ class vmmManager(vmmGObjectUI):
     #############################
 
     def vm_row_updated(self, vm):
-        row = self.rows.get(self.vm_row_key(vm), None)
-        if row is None:
+        row = self.get_row(vm)
+        if row is None:  # pragma: no cover
             return
-        self.widget("vm-list").get_model().row_changed(row.path, row.iter)
+        self.model.row_changed(row.path, row.iter)
 
     def vm_changed(self, vm):
-        row = self.rows.get(self.vm_row_key(vm), None)
+        row = self.get_row(vm)
         if row is None:
-            return
+            return  # pragma: no cover
 
         try:
             if vm == self.current_vm():
@@ -743,18 +699,18 @@ class vmmManager(vmmGObjectUI):
             row[ROW_MARKUP] = self._build_vm_markup(name, status)
 
             desc = vm.get_description()
-            row[ROW_HINT] = util.xml_escape(desc)
-        except libvirt.libvirtError, e:
-            if util.exception_is_libvirt_error(e, "VIR_ERR_NO_DOMAIN"):
+            row[ROW_HINT] = xmlutil.xml_escape(desc)
+        except Exception as e:  # pragma: no cover
+            if vm.conn.support.is_libvirt_error_no_domain(e):
                 return
             raise
 
         self.vm_row_updated(vm)
 
     def vm_inspection_changed(self, vm):
-        row = self.rows.get(self.vm_row_key(vm), None)
+        row = self.get_row(vm)
         if row is None:
-            return
+            return  # pragma: no cover
 
         new_icon = _get_inspection_icon_pixbuf(vm, 16, 16)
         row[ROW_INSPECTION_OS_ICON] = new_icon
@@ -762,30 +718,22 @@ class vmmManager(vmmGObjectUI):
         self.vm_row_updated(vm)
 
     def set_initial_selection(self, uri):
-        vmlist = self.widget("vm-list")
-        model = vmlist.get_model()
-        it = model.get_iter_first()
-        selected = None
-        while it:
-            key = model.get_value(it, ROW_HANDLE)
+        """
+        Select the passed URI in the UI. Called from engine.py via
+        cli --connect $URI
+        """
+        sel = self.widget("vm-list").get_selection()
+        for row in self.model:
+            if not row[ROW_IS_CONN]:
+                continue  # pragma: no cover
+            conn = row[ROW_HANDLE]
 
-            if key.get_uri() == uri:
-                vmlist.get_selection().select_iter(it)
+            if conn.get_uri() == uri:
+                sel.select_iter(row.iter)
                 return
 
-            if not selected:
-                vmlist.get_selection().select_iter(it)
-                selected = key
-            elif key.get_autoconnect() and not selected.get_autoconnect():
-                vmlist.get_selection().select_iter(it)
-                selected = key
-                if not uri:
-                    return
-
-            it = model.iter_next(it)
-
     def conn_state_changed(self, conn):
-        row = self.rows[conn.get_uri()]
+        row = self.get_row(conn)
         row[ROW_SORT_KEY] = conn.get_pretty_desc()
         row[ROW_MARKUP] = self._build_conn_markup(conn, row[ROW_SORT_KEY])
         row[ROW_IS_CONN_CONNECTED] = not conn.is_disconnected()
@@ -793,28 +741,19 @@ class vmmManager(vmmGObjectUI):
         row[ROW_HINT] = self._build_conn_hint(conn)
 
         if not conn.is_active():
-            # Connection went inactive, delete any VM child nodes
-            parent = row.iter
-            if parent is not None:
-                model = self.widget("vm-list").get_model()
-                child = model.iter_children(parent)
-                while child is not None:
-                    vm = model[child][ROW_HANDLE]
-                    del self.rows[self.vm_row_key(vm)]
-                    model.remove(child)
-                    child = model.iter_children(parent)
+            self._remove_child_rows(row)
 
         self.conn_row_updated(conn)
         self.update_current_selection()
 
     def conn_row_updated(self, conn):
-        row = self.rows[conn.get_uri()]
+        row = self.get_row(conn)
 
         self.max_disk_rate = max(self.max_disk_rate, conn.disk_io_max_rate())
         self.max_net_rate = max(self.max_net_rate,
                                 conn.network_traffic_max_rate())
 
-        self.widget("vm-list").get_model().row_changed(row.path, row.iter)
+        self.model.row_changed(row.path, row.iter)
 
     def change_run_text(self, can_restore):
         if can_restore:
@@ -828,10 +767,12 @@ class vmmManager(vmmGObjectUI):
 
     def update_current_selection(self, ignore=None):
         vm = self.current_vm()
+        conn = self.current_conn()
 
         show_open = bool(vm)
         show_details = bool(vm)
-        host_details = bool(len(self.rows))
+        host_details = bool(vm or conn)
+        can_delete = bool(vm or conn)
 
         show_run = bool(vm and vm.is_runable())
         is_paused = bool(vm and vm.is_paused())
@@ -858,29 +799,28 @@ class vmmManager(vmmGObjectUI):
             pauseTooltip = _("Pause the virtual machine")
         self.widget("vm-pause").set_tooltip_text(pauseTooltip)
 
+        self.widget("menu_edit_delete").set_sensitive(can_delete)
         self.widget("menu_edit_details").set_sensitive(show_details)
         self.widget("menu_host_details").set_sensitive(host_details)
 
     def popup_vm_menu_key(self, widget_ignore, event):
         if Gdk.keyval_name(event.keyval) != "Menu":
-            return False
+            return False  # pragma: no cover
 
         model, treeiter = self.widget("vm-list").get_selection().get_selected()
         self.popup_vm_menu(model, treeiter, event)
         return True
 
-    def popup_vm_menu_button(self, widget, event):
+    def popup_vm_menu_button(self, vmlist, event):
         if event.button != 3:
             return False
 
-        tup = widget.get_path_at_pos(int(event.x), int(event.y))
+        tup = vmlist.get_path_at_pos(int(event.x), int(event.y))
         if tup is None:
-            return False
+            return False  # pragma: no cover
         path = tup[0]
-        model = widget.get_model()
-        _iter = model.get_iter(path)
 
-        self.popup_vm_menu(model, _iter, event)
+        self.popup_vm_menu(self.model, self.model.get_iter(path), event)
         return False
 
     def popup_vm_menu(self, model, _iter, event):
@@ -888,7 +828,7 @@ class vmmManager(vmmGObjectUI):
             # Popup the vm menu
             vm = model[_iter][ROW_HANDLE]
             self.vmmenu.update_widget_states(vm)
-            self.vmmenu.popup(None, None, None, None, 0, event.time)
+            self.vmmenu.popup_at_pointer(event)
         else:
             # Pop up connection menu
             conn = model[_iter][ROW_HANDLE]
@@ -901,7 +841,7 @@ class vmmManager(vmmGObjectUI):
             self.connmenu_items["connect"].set_sensitive(disconn)
             self.connmenu_items["delete"].set_sensitive(disconn)
 
-            self.connmenu.popup(None, None, None, None, 0, event.time)
+            self.connmenu.popup_at_pointer(event)
 
 
     #################
@@ -911,42 +851,42 @@ class vmmManager(vmmGObjectUI):
     def vmlist_name_sorter(self, model, iter1, iter2, ignore):
         key1 = str(model[iter1][ROW_SORT_KEY]).lower()
         key2 = str(model[iter2][ROW_SORT_KEY]).lower()
-        return cmp(key1, key2)
+        return _cmp(key1, key2)
 
     def vmlist_guest_cpu_usage_sorter(self, model, iter1, iter2, ignore):
         obj1 = model[iter1][ROW_HANDLE]
         obj2 = model[iter2][ROW_HANDLE]
 
-        return cmp(obj1.guest_cpu_time_percentage(),
+        return _cmp(obj1.guest_cpu_time_percentage(),
                    obj2.guest_cpu_time_percentage())
 
     def vmlist_host_cpu_usage_sorter(self, model, iter1, iter2, ignore):
         obj1 = model[iter1][ROW_HANDLE]
         obj2 = model[iter2][ROW_HANDLE]
 
-        return cmp(obj1.host_cpu_time_percentage(),
+        return _cmp(obj1.host_cpu_time_percentage(),
                    obj2.host_cpu_time_percentage())
 
     def vmlist_memory_usage_sorter(self, model, iter1, iter2, ignore):
         obj1 = model[iter1][ROW_HANDLE]
         obj2 = model[iter2][ROW_HANDLE]
 
-        return cmp(obj1.stats_memory(),
+        return _cmp(obj1.stats_memory(),
                    obj2.stats_memory())
 
     def vmlist_disk_io_sorter(self, model, iter1, iter2, ignore):
         obj1 = model[iter1][ROW_HANDLE]
         obj2 = model[iter2][ROW_HANDLE]
 
-        return cmp(obj1.disk_io_rate(), obj2.disk_io_rate())
+        return _cmp(obj1.disk_io_rate(), obj2.disk_io_rate())
 
     def vmlist_network_usage_sorter(self, model, iter1, iter2, ignore):
         obj1 = model[iter1][ROW_HANDLE]
         obj2 = model[iter2][ROW_HANDLE]
 
-        return cmp(obj1.network_traffic_rate(), obj2.network_traffic_rate())
+        return _cmp(obj1.network_traffic_rate(), obj2.network_traffic_rate())
 
-    def enable_polling(self, column):
+    def _config_polling_change_cb(self, column):
         # pylint: disable=redefined-variable-type
         if column == COL_GUEST_CPU:
             widgn = ["menu_view_stats_guest_cpu", "menu_view_stats_host_cpu"]
@@ -961,15 +901,14 @@ class vmmManager(vmmGObjectUI):
             widgn = "menu_view_stats_memory"
             do_enable = self.config.get_stats_enable_memory_poll()
 
-        for w in util.listify(widgn):
+        for w in xmlutil.listify(widgn):
             widget = self.widget(w)
             tool_text = ""
 
             if do_enable:
                 widget.set_sensitive(True)
             else:
-                if widget.get_active():
-                    widget.set_active(False)
+                widget.set_active(False)
                 widget.set_sensitive(False)
                 tool_text = _("Disabled in preferences dialog.")
             widget.set_tooltip_text(tool_text)
