@@ -12,6 +12,7 @@ from virtinst import log
 from . import vmmenu
 from .baseclass import vmmGObjectUI
 from .engine import vmmEngine
+from .details.console import vmmConsolePages
 from .details.details import vmmDetails
 from .details.snapshots import vmmSnapshotPage
 
@@ -38,7 +39,7 @@ class vmmVMWindow(vmmGObjectUI):
             if key not in cls._instances:
                 cls._instances[key] = vmmVMWindow(vm)
             return cls._instances[key]
-        except Exception as e:
+        except Exception as e:  # pragma: no cover
             if not parentobj:
                 raise
             parentobj.err.show_err(
@@ -63,27 +64,31 @@ class vmmVMWindow(vmmGObjectUI):
             pages = self.widget("details-pages")
             pages.set_current_page(DETAILS_PAGE_DETAILS)
         else:
-            self.conn.connect("vm-removed", self._vm_removed)
-
-        self._mediacombo = None
+            self.conn.connect("vm-removed", self._vm_removed_cb)
 
         self.ignoreDetails = False
 
-        from .details.console import vmmConsolePages
-        self.console = vmmConsolePages(self.vm, self.builder, self.topwin)
-        self.snapshots = vmmSnapshotPage(self.vm, self.builder, self.topwin)
-        self.widget("snapshot-placeholder").add(self.snapshots.top_box)
+        self._console = vmmConsolePages(self.vm, self.builder, self.topwin)
+        self.widget("console-placeholder").add(self._console.top_box)
+        self._console.connect("page-changed", self._console_page_changed_cb)
+        self._console.connect("leave-fullscreen",
+                self._console_leave_fullscreen_cb)
+        self._console.connect("change-title",
+                self._console_change_title_cb)
+
+        self._snapshots = vmmSnapshotPage(self.vm, self.builder, self.topwin)
+        self.widget("snapshot-placeholder").add(self._snapshots.top_box)
+
         self._details = vmmDetails(self.vm, self.builder, self.topwin,
                 self.is_customize_dialog)
         self.widget("details-placeholder").add(self._details.top_box)
 
         # Set default window size
         w, h = self.vm.get_details_window_size()
-        if w <= 0:
-            w = 800
-        if h <= 0:
-            h = 600
-        self.topwin.set_default_size(w, h)
+        if w <= 0 or h <= 0:
+            self._set_initial_window_size()
+        else:
+            self.topwin.set_default_size(w, h)
         self._window_size = None
 
         self._shutdownmenu = None
@@ -117,37 +122,39 @@ class vmmVMWindow(vmmGObjectUI):
             "on_details_menu_view_console_toggled": self.details_console_changed,
             "on_details_menu_view_snapshots_toggled": self.details_console_changed,
 
-            "on_details_pages_switch_page": self.switch_page,
+            "on_details_pages_switch_page": self._details_page_switch_cb,
 
-            # Listeners stored in vmmConsolePages
-            "on_details_menu_view_fullscreen_activate": (
-                self.console.details_toggle_fullscreen),
-            "on_details_menu_view_size_to_vm_activate": (
-                self.console.details_size_to_vm),
-            "on_details_menu_view_scale_always_toggled": (
-                self.console.details_scaling_ui_changed_cb),
-            "on_details_menu_view_scale_fullscreen_toggled": (
-                self.console.details_scaling_ui_changed_cb),
-            "on_details_menu_view_scale_never_toggled": (
-                self.console.details_scaling_ui_changed_cb),
-            "on_details_menu_view_resizeguest_toggled": (
-                self.console.details_resizeguest_ui_changed_cb),
-
-            "on_console_pages_switch_page": (
-                self.console.details_page_changed),
-            "on_console_auth_password_activate": (
-                self.console.details_auth_login),
-            "on_console_auth_login_clicked": (
-                self.console.details_auth_login),
+            "on_details_menu_view_fullscreen_activate": self._fullscreen_changed_cb,
+            "on_details_menu_view_size_to_vm_activate": self._size_to_vm_cb,
+            "on_details_menu_view_scale_always_toggled": self._scaling_ui_changed_cb,
+            "on_details_menu_view_scale_fullscreen_toggled": self._scaling_ui_changed_cb,
+            "on_details_menu_view_scale_never_toggled": self._scaling_ui_changed_cb,
+            "on_details_menu_view_resizeguest_toggled": self._resizeguest_ui_changed_cb,
+            "on_details_menu_view_autoconnect_activate": self._autoconnect_ui_changed_cb,
         })
 
         # Deliberately keep all this after signal connection
-        self.vm.connect("state-changed", self.refresh_vm_state)
-        self.vm.connect("resources-sampled", self.refresh_resources)
-        self.vm.connect("inspection-changed",
-                lambda *x: self._details.refresh_os_page())
+        self.vm.connect("state-changed", self._vm_state_changed_cb)
+        self.vm.connect("resources-sampled", self._resources_sampled_cb)
 
-        self.refresh_vm_state()
+        self._sync_console_page_menu_state()
+        self._console_refresh_scaling_from_settings()
+
+        self.add_gsettings_handle(
+            self.vm.on_console_scaling_changed(
+                self._console_refresh_scaling_from_settings))
+
+        self._console_refresh_resizeguest_from_settings()
+        self.add_gsettings_handle(
+            self.vm.on_console_resizeguest_changed(
+                self._console_refresh_resizeguest_from_settings))
+
+        self._console_refresh_autoconnect_from_settings()
+        self.add_gsettings_handle(
+            self.vm.on_console_autoconnect_changed(
+                self._console_refresh_autoconnect_from_settings))
+
+        self._refresh_vm_state()
         self.activate_default_page()
 
 
@@ -156,10 +163,10 @@ class vmmVMWindow(vmmGObjectUI):
         return self.vm.conn
 
     def _cleanup(self):
-        self.console.cleanup()
-        self.console = None
-        self.snapshots.cleanup()
-        self.snapshots = None
+        self._console.cleanup()
+        self._console = None
+        self._snapshots.cleanup()
+        self._snapshots = None
         self._details.cleanup()
         self._details = None
         self._shutdownmenu.destroy()
@@ -181,7 +188,7 @@ class vmmVMWindow(vmmGObjectUI):
             return
 
         vmmEngine.get_instance().increment_window_counter()
-        self.refresh_vm_state()
+        self._refresh_vm_state()
 
     def customize_finish(self, src):
         ignore = src
@@ -189,8 +196,26 @@ class vmmVMWindow(vmmGObjectUI):
             return
         self.emit("customize-finished", self.vm)
 
-    def _vm_removed(self, _conn, connkey):
-        if self.vm.get_connkey() == connkey:
+    def _set_initial_window_size(self):
+        """
+        We want the window size for new windows to be 1024x768 viewer
+        size, plus whatever it takes to fit the toolbar+menubar, etc.
+        To achieve this, we force the display box to the desired size
+        with set_size_request, wait for the window to report it has
+        been resized, and then unset the hardcoded size request so
+        the user can manually resize the window however they want.
+        """
+        w = 1024
+        h = 768
+        hid = []
+        def win_cb(src, event):
+            self.widget("details-pages").set_size_request(-1, -1)
+            self.topwin.disconnect(hid[0])
+        self.widget("details-pages").set_size_request(w, h)
+        hid.append(self.topwin.connect("configure-event", win_cb))
+
+    def _vm_removed_cb(self, _conn, vm):
+        if self.vm == vm:
             self.cleanup()
 
     def _customize_cancel(self):
@@ -220,17 +245,14 @@ class vmmVMWindow(vmmGObjectUI):
     def _close(self):
         fs = self.widget("details-menu-view-fullscreen")
         if fs.get_active():
-            fs.set_active(False)
+            fs.set_active(False)  # pragma: no cover
 
         if not self.is_visible():
             return
 
         self.topwin.hide()
-        if self.console.details_viewer_is_visible():
-            try:
-                self.console.details_close_viewer()
-            except Exception:
-                log.error("Failure when disconnecting from desktop server")
+        self._console.vmwindow_close()
+        self._details.vmwindow_close()
 
         self.emit("closed")
         vmmEngine.get_instance().decrement_window_counter()
@@ -261,6 +283,14 @@ class vmmVMWindow(vmmGObjectUI):
         self.widget("details-menu-view-toolbar").set_active(
                                     self.config.get_details_show_toolbar())
 
+        # Keycombo menu (ctrl+alt+del etc.)
+        self.widget("details-menu-send-key").set_submenu(
+                self._console.vmwindow_get_keycombo_menu())
+
+        # Serial list menu
+        self.widget("details-menu-view-console-list").set_submenu(
+                self._console.vmwindow_get_console_list_menu())
+
 
     ##########################
     # Window state listeners #
@@ -268,7 +298,7 @@ class vmmVMWindow(vmmGObjectUI):
 
     def window_resized(self, ignore, ignore2):
         if not self.is_visible():
-            return
+            return  # pragma: no cover
         self._window_size = self.topwin.get_size()
 
     def control_fullscreen(self, src):
@@ -282,12 +312,8 @@ class vmmVMWindow(vmmGObjectUI):
 
         active = src.get_active()
         self.config.set_details_show_toolbar(active)
-
-        if (active and not
-            self.widget("details-menu-view-fullscreen").get_active()):
-            self.widget("toolbar-box").show()
-        else:
-            self.widget("toolbar-box").hide()
+        fsactive = self.widget("details-menu-view-fullscreen").get_active()
+        self.widget("toolbar-box").set_visible(active and not fsactive)
 
     def details_console_changed(self, src):
         if self.ignoreDetails:
@@ -304,19 +330,17 @@ class vmmVMWindow(vmmGObjectUI):
         pages = self.widget("details-pages")
         if pages.get_current_page() == DETAILS_PAGE_DETAILS:
             if self._details.vmwindow_has_unapplied_changes():
-                self.sync_details_console_view(True)
+                self._sync_toolbar_page_buttons(pages.get_current_page())
                 return
-            self._details.disable_apply()
 
         if is_details:
             pages.set_current_page(DETAILS_PAGE_DETAILS)
         elif is_snapshot:
-            self.snapshots.show_page()
             pages.set_current_page(DETAILS_PAGE_SNAPSHOTS)
         else:
             pages.set_current_page(DETAILS_PAGE_CONSOLE)
 
-    def sync_details_console_view(self, newpage):
+    def _sync_toolbar_page_buttons(self, newpage):
         details = self.widget("control-vm-details")
         details_menu = self.widget("details-menu-view-details")
         console = self.widget("control-vm-console")
@@ -340,15 +364,14 @@ class vmmVMWindow(vmmGObjectUI):
         finally:
             self.ignoreDetails = False
 
-    def switch_page(self, notebook=None, ignore2=None, newpage=None):
+    def _details_page_switch_cb(self, notebook, pagewidget, newpage):
         for i in range(notebook.get_n_pages()):
             w = notebook.get_nth_page(i)
             w.set_visible(i == newpage)
 
-        self.page_refresh(newpage)
-
-        self.sync_details_console_view(newpage)
-        self.console.details_refresh_can_fullscreen()
+        self._refresh_current_page(newpage)
+        self._sync_toolbar_page_buttons(newpage)
+        self._sync_console_page_menu_state()
 
     def change_run_text(self, can_restore):
         if can_restore:
@@ -360,8 +383,21 @@ class vmmVMWindow(vmmGObjectUI):
         self.widget("details-vm-menu").get_submenu().change_run_text(text)
         self.widget("control-run").set_label(strip_text)
 
-    def refresh_vm_state(self, ignore=None):
+    def _refresh_title(self):
+        title = (_("%(vm-name)s on %(connection-name)s") % {
+            "vm-name": self.vm.get_name_or_title(),
+            "connection-name": self.vm.conn.get_pretty_desc(),
+        })
+
+        grabmsg = self._console.vmwindow_get_title_message()
+        if grabmsg:
+            title = grabmsg + " " + title
+
+        self.topwin.set_title(title)
+
+    def _refresh_vm_state(self):
         vm = self.vm
+        self._refresh_title()
 
         self.widget("details-menu-view-toolbar").set_active(
             self.config.get_details_show_toolbar())
@@ -397,13 +433,7 @@ class vmmVMWindow(vmmGObjectUI):
             tooltip += "\n" + errmsg
         self.widget("control-snapshots").set_tooltip_text(tooltip)
 
-        details = self.widget("details-pages")
-        self.page_refresh(details.get_current_page())
-
-        self._details.vmwindow_refresh_vm_state()
-        self.console.details_update_widget_states()
-        if not run:
-            self.activate_default_console_page()
+        self._refresh_current_page()
 
 
     #############################
@@ -418,15 +448,7 @@ class vmmVMWindow(vmmGObjectUI):
         vmmEngine.get_instance().exit_app()
 
     def activate_default_console_page(self):
-        pages = self.widget("details-pages")
-
-        # console.activate_default_console_page() will as a side effect
-        # switch to DETAILS_PAGE_CONSOLE. However this code path is triggered
-        # when the user runs a VM while they are focused on the details page,
-        # and we don't want to switch pages out from under them.
-        origpage = pages.get_current_page()
-        self.console.details_activate_default_console_page()
-        pages.set_current_page(origpage)
+        self._console.vmwindow_activate_default_console_page()
 
     # activate_* are called from engine.py via CLI options
     def activate_default_page(self):
@@ -468,8 +490,8 @@ class vmmVMWindow(vmmGObjectUI):
             vmmenu.VMActionUI.resume(self, self.vm)
 
     def control_vm_menu(self, src_ignore):
-        can_usb = bool(self.console.details_viewer_has_usb_redirection() and
-                       self.vm.has_spicevmc_type_redirdev())
+        can_usb = bool(self.vm.has_spicevmc_type_redirdev() and
+                       self._console.vmwindow_viewer_has_usb_redirection())
         self.widget("details-menu-usb-redirection").set_sensitive(can_usb)
 
     def control_vm_run(self, src_ignore):
@@ -484,15 +506,15 @@ class vmmVMWindow(vmmGObjectUI):
         ignore = src
         try:
             return self._take_screenshot()
-        except Exception as e:
+        except Exception as e:  # pragma: no cover
             self.err.show_err(_("Error taking screenshot: %s") % str(e))
 
     def control_vm_usb_redirection(self, src):
         ignore = src
         spice_usbdev_dialog = self.err
 
-        spice_usbdev_widget = self.console.details_viewer_get_usb_widget()
-        if not spice_usbdev_widget:
+        spice_usbdev_widget = self._console.vmwindow_viewer_get_usb_widget()
+        if not spice_usbdev_widget:  # pragma: no cover
             self.err.show_err(_("Error initializing spice USB device widget"))
             return
 
@@ -502,7 +524,7 @@ class vmmVMWindow(vmmGObjectUI):
                                       buttons=Gtk.ButtonsType.CLOSE)
 
     def _take_screenshot(self):
-        image = self.console.details_viewer_get_pixbuf()
+        image = self._console.vmwindow_viewer_get_pixbuf()
 
         metadata = {
             'tEXt::Hypervisor URI': self.vm.conn.get_uri(),
@@ -522,7 +544,7 @@ class vmmVMWindow(vmmGObjectUI):
             ret = ret[1]
         # F24 rawhide, ret[1] is a named tuple with a 'buffer' element...
         if hasattr(ret, "buffer"):
-            ret = ret.buffer
+            ret = ret.buffer  # pragma: no cover
 
         import datetime
         now = str(datetime.datetime.now()).split(".")[0].replace(" ", "_")
@@ -534,13 +556,13 @@ class vmmVMWindow(vmmGObjectUI):
             dialog_type=Gtk.FileChooserAction.SAVE,
             browse_reason=self.config.CONFIG_DIR_SCREENSHOT,
             default_name=default)
-        if not path:
+        if not path:  # pragma: no cover
             log.debug("No screenshot path given, skipping save.")
             return
 
         filename = path
         if not filename.endswith(".png"):
-            filename += ".png"
+            filename += ".png"  # pragma: no cover
         open(filename, "wb").write(ret)
 
 
@@ -548,24 +570,139 @@ class vmmVMWindow(vmmGObjectUI):
     # Details page refresh #
     ########################
 
-    def refresh_resources(self, ignore):
+    def _refresh_resources(self):
         details = self.widget("details-pages")
         page = details.get_current_page()
 
-        # If the dialog is visible, we want to make sure the XML is always
-        # up to date
-        try:
-            if self.is_visible():
-                self.vm.ensure_latest_xml()
-        except Exception as e:
-            if self.conn.support.is_libvirt_error_no_domain(e):
-                self.close()
-                return
-            raise
-
         if page == DETAILS_PAGE_DETAILS:
             self._details.vmwindow_resources_refreshed()
 
-    def page_refresh(self, page):
-        if page == DETAILS_PAGE_DETAILS:
-            self._details.vmwindow_page_refresh()
+    def _refresh_current_page(self, newpage=None):
+        newpage = newpage or self.widget("details-pages").get_current_page()
+
+        is_details = newpage == DETAILS_PAGE_DETAILS
+        self._details.vmwindow_refresh_vm_state(is_details)
+
+        if newpage == DETAILS_PAGE_CONSOLE:
+            self._console.vmwindow_refresh_vm_state()
+        elif newpage == DETAILS_PAGE_SNAPSHOTS:
+            self._snapshots.vmwindow_refresh_vm_state()
+
+
+    #########################
+    # Console page handling #
+    #########################
+
+    def _sync_console_page_menu_state(self):
+        if not self.vm:
+            # This is triggered via cleanup + idle_add, so vm might
+            # disappear and spam the logs
+            return  # pragma: no cover
+
+        paused = self.vm.is_paused()
+        is_viewer = self._console.vmwindow_get_viewer_is_visible()
+        can_usb = self._console.vmwindow_get_can_usb_redirect()
+
+        self.widget("details-menu-vm-screenshot").set_sensitive(is_viewer)
+        self.widget("details-menu-usb-redirection").set_sensitive(can_usb)
+        keycombo_menu = self._console.vmwindow_get_keycombo_menu()
+
+        can_sendkey = (is_viewer and not paused)
+        for c in keycombo_menu.get_children():
+            c.set_sensitive(can_sendkey)
+
+        self._console_refresh_can_fullscreen()
+        self._console_refresh_resizeguest_from_settings()
+
+    def _console_refresh_can_fullscreen(self):
+        allow_fullscreen = self._console.vmwindow_get_viewer_is_visible()
+
+        self.widget("control-fullscreen").set_sensitive(allow_fullscreen)
+        self.widget("details-menu-view-fullscreen").set_sensitive(
+            allow_fullscreen)
+
+    def _console_refresh_scaling_from_settings(self):
+        scale_type = self.vm.get_console_scaling()
+        self.widget("details-menu-view-scale-always").set_active(
+            scale_type == self.config.CONSOLE_SCALE_ALWAYS)
+        self.widget("details-menu-view-scale-never").set_active(
+            scale_type == self.config.CONSOLE_SCALE_NEVER)
+        self.widget("details-menu-view-scale-fullscreen").set_active(
+            scale_type == self.config.CONSOLE_SCALE_FULLSCREEN)
+
+        self._console.vmwindow_sync_scaling_with_display()
+
+    def _scaling_ui_changed_cb(self, src):
+        # Called from details.py
+        if not src.get_active():
+            return
+
+        scale_type = 0
+        if src == self.widget("details-menu-view-scale-always"):
+            scale_type = self.config.CONSOLE_SCALE_ALWAYS
+        elif src == self.widget("details-menu-view-scale-fullscreen"):
+            scale_type = self.config.CONSOLE_SCALE_FULLSCREEN
+        elif src == self.widget("details-menu-view-scale-never"):
+            scale_type = self.config.CONSOLE_SCALE_NEVER
+
+        self.vm.set_console_scaling(scale_type)
+
+    def _fullscreen_changed_cb(self, src):
+        do_fullscreen = src.get_active()
+        self.widget("control-fullscreen").set_active(do_fullscreen)
+        self._console.vmwindow_set_fullscreen(do_fullscreen)
+
+        self.widget("details-menubar").set_visible(not do_fullscreen)
+
+        show_toolbar = not do_fullscreen
+        if not self.widget("details-menu-view-toolbar").get_active():
+            show_toolbar = False  # pragma: no cover
+        self.widget("toolbar-box").set_visible(show_toolbar)
+
+    def _resizeguest_ui_changed_cb(self, src):
+        if not src.get_sensitive():
+            return  # pragma: no cover
+
+        val = int(self.widget("details-menu-view-resizeguest").get_active())
+        self.vm.set_console_resizeguest(val)
+        self._console.vmwindow_sync_resizeguest_with_display()
+
+    def _console_refresh_resizeguest_from_settings(self):
+        tooltip = self._console.vmwindow_get_resizeguest_tooltip()
+        val = self.vm.get_console_resizeguest()
+        widget = self.widget("details-menu-view-resizeguest")
+        widget.set_tooltip_text(tooltip)
+        widget.set_sensitive(not bool(tooltip))
+        if not tooltip:
+            self.widget("details-menu-view-resizeguest").set_active(bool(val))
+
+        self._console.vmwindow_sync_resizeguest_with_display()
+
+    def _autoconnect_ui_changed_cb(self, src):
+        val = int(self.widget("details-menu-view-autoconnect").get_active())
+        self.vm.set_console_autoconnect(val)
+
+    def _console_refresh_autoconnect_from_settings(self):
+        val = self.vm.get_console_autoconnect()
+        self.widget("details-menu-view-autoconnect").set_active(val)
+
+    def _size_to_vm_cb(self, src):
+        self._console.vmwindow_set_size_to_vm()
+
+    def _console_leave_fullscreen_cb(self, src):
+        # This will trigger de-fullscreening in a roundabout way
+        self.widget("control-fullscreen").set_active(False)
+
+    def _console_change_title_cb(self, src):
+        self._refresh_title()
+
+    def _vm_state_changed_cb(self, src):
+        if self.is_visible():
+            self._refresh_vm_state()
+
+    def _resources_sampled_cb(self, src):
+        if self.is_visible():
+            self._refresh_resources()
+
+    def _console_page_changed_cb(self, src):
+        self._sync_console_page_menu_state()
