@@ -3,47 +3,19 @@
 #
 # Copyright 2013 Red Hat, Inc.
 #
-# This program is free software; you can redistribute it and/or modify
-# it under the terms of the GNU General Public License as published by
-# the Free Software Foundation; either version 2 of the License, or
-# (at your option) any later version.
-#
-# This program is distributed in the hope that it will be useful,
-# but WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-# GNU General Public License for more details.
-#
-# You should have received a copy of the GNU General Public License
-# along with this program; if not, write to the Free Software
-# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
-# MA 02110-1301 USA.
+# This work is licensed under the GNU GPLv2 or later.
+# See the COPYING file in the top-level directory.
 
-import logging
 import os
 import re
 import stat
-import statvfs
+import subprocess
 
 import libvirt
 
+from .logger import log
 from .storage import StoragePool, StorageVolume
-
-
-def _lookup_pool_by_dirname(conn, path):
-    """
-    Try to find the parent pool for the passed path.
-    If found, and the pool isn't running, attempt to start it up.
-
-    return pool, or None if not found
-    """
-    pool = StoragePool.lookup_pool_by_path(conn, os.path.dirname(path))
-    if not pool:
-        return None
-
-    # Ensure pool is running
-    if pool.info()[0] != libvirt.VIR_STORAGE_POOL_RUNNING:
-        pool.create(0)
-    return pool
+from . import xmlutil
 
 
 def _lookup_vol_by_path(conn, path):
@@ -56,9 +28,14 @@ def _lookup_vol_by_path(conn, path):
         vol.info()
         return vol, None
     except libvirt.libvirtError as e:
-        if (hasattr(libvirt, "VIR_ERR_NO_STORAGE_VOL") and
+        # test_urls trigger empty errors here, because python
+        # garbage collection kicks in after the failure but before
+        # we read the error code, and libvirt virStoragePoolFree
+        # public entry point clears the cached error. So ignore
+        # an empty error code
+        if (e.get_error_code() and
             e.get_error_code() != libvirt.VIR_ERR_NO_STORAGE_VOL):
-            raise
+            raise  # pragma: no cover
         return None, e
 
 
@@ -73,32 +50,32 @@ def _lookup_vol_by_basename(pool, path):
         return pool.storageVolLookupByName(name)
 
 
-def _stat_disk(path):
-    """
-    Returns the tuple (isreg, size)
-    """
+def _get_block_size(path):  # pragma: no cover
+    try:
+        fd = os.open(path, os.O_RDONLY)
+        # os.SEEK_END is not present on all systems
+        size = os.lseek(fd, 0, 2)
+        os.close(fd)
+    except Exception:
+        size = 0
+    return size
+
+
+def _get_size(path):
     if not os.path.exists(path):
-        return True, 0
-
-    mode = os.stat(path)[stat.ST_MODE]
-
-    # os.path.getsize('/dev/..') can be zero on some platforms
-    if stat.S_ISBLK(mode):
-        try:
-            fd = os.open(path, os.O_RDONLY)
-            # os.SEEK_END is not present on all systems
-            size = os.lseek(fd, 0, 2)
-            os.close(fd)
-        except Exception:
-            size = 0
-        return False, size
-    elif stat.S_ISREG(mode):
-        return True, os.path.getsize(path)
-
-    return True, 0
+        return 0
+    if _stat_is_block(path):
+        return _get_block_size(path)  # pragma: no cover
+    return os.path.getsize(path)
 
 
-def check_if_path_managed(conn, path):
+def _stat_is_block(path):
+    if not os.path.exists(path):
+        return False
+    return stat.S_ISBLK(os.stat(path)[stat.ST_MODE])
+
+
+def _check_if_path_managed(conn, path):
     """
     Try to lookup storage objects for the passed path.
 
@@ -108,27 +85,27 @@ def check_if_path_managed(conn, path):
     if vol:
         return vol, vol.storagePoolLookupByVolume()
 
-    pool = _lookup_pool_by_dirname(conn, path)
+    pool = StoragePool.lookup_pool_by_path(conn, os.path.dirname(path))
     if not pool:
         return None, None
 
     # We have the parent pool, but didn't find a volume on first lookup
-    # attempt. Refresh the pool and try again, incase we were just out
-    # of date.
+    # attempt. Refresh the pool and try again, in case we were just out
+    # of date or the pool was inactive.
     try:
-        pool.refresh(0)
+        StoragePool.ensure_pool_is_running(pool, refresh=True)
         vol, verr = _lookup_vol_by_path(conn, path)
         if verr:
             try:
                 vol = _lookup_vol_by_basename(pool, path)
-            except Exception:
+            except Exception:  # pragma: no cover
                 pass
-    except Exception as e:
+    except Exception as e:  # pragma: no cover
         vol = None
         pool = None
         verr = str(e)
 
-    if not vol and not pool and verr:
+    if not vol and not pool and verr:  # pragma: no cover
         raise ValueError(_("Cannot use storage %(path)s: %(err)s") %
             {'path': path, 'err': verr})
 
@@ -148,18 +125,31 @@ def _can_auto_manage(path):
     return True
 
 
+def _get_storage_search_path(path):
+    # If the passed path is one of our artificial rbd:// style
+    # URIs, parse out the path component, since that is what is needed
+    # for looking up storage volumes by target path
+    from .uri import URI
+    uriobj = URI(path)
+    if uriobj.scheme == "rbd":
+        return uriobj.path.strip("/")
+    return path
+
+
 def manage_path(conn, path):
     """
     If path is not managed, try to create a storage pool to probe the path
     """
-    if not conn.check_support(conn.SUPPORT_CONN_STORAGE):
-        return None, None
+    if not conn.support.conn_storage():
+        return None, None  # pragma: no cover
     if not path:
         return None, None
 
-    if not path_is_url(path):
+    if not path_is_url(path) and not path_is_network_vol(conn, path):
         path = os.path.abspath(path)
-    vol, pool = check_if_path_managed(conn, path)
+
+    searchpath = _get_storage_search_path(path)
+    vol, pool = _check_if_path_managed(conn, searchpath)
     if vol or pool or not _can_auto_manage(path):
         return vol, pool
 
@@ -168,7 +158,7 @@ def manage_path(conn, path):
     if not poolname:
         poolname = "dirpool"
     poolname = StoragePool.find_free_name(conn, poolname)
-    logging.debug("Attempting to build pool=%s target=%s", poolname, dirname)
+    log.debug("Attempting to build pool=%s target=%s", poolname, dirname)
 
     poolxml = StoragePool(conn)
     poolxml.name = poolname
@@ -184,34 +174,25 @@ def path_is_url(path):
     """
     Detect if path is a URL
     """
-    if not path:
-        return False
-    return bool(re.match("[a-zA-Z]+(\+[a-zA-Z]+)?://.*", path))
+    return bool(re.match(r"[a-zA-Z]+(\+[a-zA-Z]+)?://.*", path or ""))
+
+
+def path_is_network_vol(conn, path):
+    """
+    Detect if path is a network volume such as rbd, gluster, etc
+    """
+    for volxml in conn.fetch_all_vols():
+        if path and volxml.target_path == path:
+            return volxml.type == "network"
+    return False
 
 
 def _get_dev_type(path, vol_xml, vol_object, pool_xml, remote):
     """
     Try to get device type for volume.
     """
-    if vol_xml:
-        if vol_xml.type:
-            return vol_xml.type
-
-        # If vol_xml.type is None the vol_xml.file_type can return only
-        # these types: block, network or file
-        if vol_xml.file_type == libvirt.VIR_STORAGE_VOL_BLOCK:
-            return "block"
-        elif vol_xml.file_type == libvirt.VIR_STORAGE_VOL_NETWORK:
-            return "network"
-
-    if vol_object:
-        t = vol_object.info()[0]
-        if t == StorageVolume.TYPE_FILE:
-            return "file"
-        elif t == StorageVolume.TYPE_BLOCK:
-            return "block"
-        elif t == StorageVolume.TYPE_NETWORK:
-            return "network"
+    if vol_xml and vol_xml.type:
+        return vol_xml.type
 
     if pool_xml:
         t = pool_xml.get_disk_type()
@@ -220,28 +201,181 @@ def _get_dev_type(path, vol_xml, vol_object, pool_xml, remote):
         elif t == StorageVolume.TYPE_NETWORK:
             return "network"
 
+    if vol_object:  # pragma: no cover
+        # This path is hard to test, because test suite XML always has
+        # the vol_xml.type set
+        t = vol_object.info()[0]
+        if t == StorageVolume.TYPE_FILE:
+            return "file"
+        elif t == StorageVolume.TYPE_BLOCK:
+            return "block"
+        elif t == StorageVolume.TYPE_NETWORK:
+            return "network"
+
     if path:
         if path_is_url(path):
             return "network"
 
-        if not remote:
-            if os.path.isdir(path):
-                return "dir"
-            elif _stat_disk(path)[0]:
-                return "file"
-            else:
+        if remote:
+            if not _can_auto_manage(path):
+                # Just a heurisitic, if this path is one of the ones
+                # we don't try to auto-import, then consider it a
+                # block device, because managing those correctly is difficult
                 return "block"
 
+        else:
+            if os.path.isdir(path):
+                return "dir"
+            elif _stat_is_block(path):
+                return "block"  # pragma: no cover
+
     return "file"
 
 
+def path_definitely_exists(conn, path):
+    """
+    Return True if the path certainly exists, False if we are unsure.
+    See DeviceDisk entry point for more details
+    """
+    if path is None:
+        return False
+
+    try:
+        (vol, pool) = _check_if_path_managed(conn, path)
+        ignore = pool
+        if vol:
+            return True
+
+        if not conn.is_remote():
+            return os.path.exists(path)
+    except Exception:  # pragma: no cover
+        pass
+
+    return False
+
+
+#########################
+# ACL/path perm helpers #
+#########################
+
+SETFACL = "setfacl"
+
+
+def _fix_perms_acl(dirname, username):
+    cmd = [SETFACL, "--modify", "user:%s:x" % username, dirname]
+    proc = subprocess.Popen(cmd,
+                            stdout=subprocess.PIPE,
+                            stderr=subprocess.PIPE)
+    out, err = proc.communicate()
+
+    log.debug("Ran command '%s'", cmd)
+    if out or err:
+        log.debug("out=%s\nerr=%s", out, err)
+
+    if proc.returncode != 0:
+        raise ValueError(err)
+
+
+def _fix_perms_chmod(dirname):
+    log.debug("Setting +x on %s", dirname)
+    mode = os.stat(dirname).st_mode
+    newmode = mode | stat.S_IXOTH
+    os.chmod(dirname, newmode)
+    if os.stat(dirname).st_mode != newmode:
+        # Trying to change perms on vfat at least doesn't work
+        # but also doesn't seem to error. Try and detect that
+        raise ValueError(  # pragma: no cover
+                _("Permissions on '%s' did not stick") % dirname)
+
+
+def set_dirs_searchable(dirlist, username):
+    useacl = True
+    errdict = {}
+    for dirname in dirlist:
+        if useacl:
+            try:
+                _fix_perms_acl(dirname, username)
+                continue
+            except Exception as e:
+                log.debug("setfacl failed: %s", e)
+                log.debug("trying chmod")
+                useacl = False
+
+        try:
+            # If we reach here, ACL setting failed, try chmod
+            _fix_perms_chmod(dirname)
+        except Exception as e:  # pragma: no cover
+            errdict[dirname] = str(e)
+
+    return errdict
+
+
+def _is_dir_searchable(dirname, uid, username):
+    """
+    Check if passed directory is searchable by uid
+    """
+    try:
+        statinfo = os.stat(dirname)
+    except OSError:  # pragma: no cover
+        return False
+
+    if uid == statinfo.st_uid:
+        flag = stat.S_IXUSR
+    elif uid == statinfo.st_gid:
+        flag = stat.S_IXGRP  # pragma: no cover
+    else:
+        flag = stat.S_IXOTH
+
+    if bool(statinfo.st_mode & flag):
+        return True
+
+    # Check POSIX ACL (since that is what we use to 'fix' access)
+    cmd = ["getfacl", dirname]
+    try:
+        proc = subprocess.Popen(cmd,
+                                stdout=subprocess.PIPE,
+                                stderr=subprocess.PIPE)
+        out, err = proc.communicate()
+    except OSError:  # pragma: no cover
+        log.debug("Didn't find the getfacl command.")
+        return False
+
+    if proc.returncode != 0:  # pragma: no cover
+        log.debug("Cmd '%s' failed: %s", cmd, err)
+        return False
+
+    pattern = "user:%s:..x" % username
+    return bool(re.search(pattern.encode("utf-8", "replace"), out))
+
+
+def is_path_searchable(path, uid, username):
+    """
+    Check each dir component of the passed path, see if they are
+    searchable by the uid/username, and return a list of paths
+    which aren't searchable
+    """
+    if os.path.isdir(path):
+        dirname = path
+        base = "-"
+    else:
+        dirname, base = os.path.split(path)
+
+    fixlist = []
+    while base:
+        if not _is_dir_searchable(dirname, uid, username):
+            fixlist.append(dirname)
+        dirname, base = os.path.split(dirname)
+
+    return fixlist
+
+
 ##############################################
 # Classes for tracking storage media details #
 ##############################################
 
 class _StorageBase(object):
     """
-    Storage base class, defining the API used by VirtualDisk
+    Storage base class, defining the API used by DeviceDisk
     """
     def __init__(self, conn):
         self._conn = conn
@@ -264,19 +398,24 @@ class _StorageBase(object):
             self._parent_pool_xml = StoragePool(self._conn,
                 parsexml=self.get_parent_pool().XMLDesc(0))
         return self._parent_pool_xml
-    def validate(self, disk):
+    def validate(self):
         raise NotImplementedError()
     def get_path(self):
         raise NotImplementedError()
+    def is_stub(self):
+        return False
 
     # Storage creation routines
     def is_size_conflict(self):
         raise NotImplementedError()
-    def create(self, progresscb):
-        raise NotImplementedError()
     def will_create_storage(self):
         raise NotImplementedError()
 
+    def create(self, progresscb):
+        ignore = progresscb  # pragma: no cover
+        raise xmlutil.DevError(
+            "%s can't create storage" % self.__class__.__name__)
+
 
 class _StorageCreator(_StorageBase):
     """
@@ -297,7 +436,11 @@ class _StorageCreator(_StorageBase):
     ##############
 
     def create(self, progresscb):
-        raise NotImplementedError()
+        raise NotImplementedError
+    def validate(self):
+        raise NotImplementedError
+    def get_size(self):
+        raise NotImplementedError
 
     def get_path(self):
         if self._vol_install and not self._path:
@@ -306,11 +449,8 @@ class _StorageCreator(_StorageBase):
             if self.get_dev_type() == "network":
                 self._path = self._vol_install.name
             else:
-                sep = "/"
-                if xmlobj.target_path == "" or xmlobj.target_path[-1] == '/':
-                    sep = ""
-                self._path = (xmlobj.target_path + sep +
-                              self._vol_install.name)
+                self._path = os.path.join(
+                        xmlobj.target_path, self._vol_install.name)
         return self._path
 
     def get_vol_install(self):
@@ -318,12 +458,6 @@ class _StorageCreator(_StorageBase):
     def get_vol_xml(self):
         return self._vol_install
 
-    def get_size(self):
-        if self._size is None:
-            self._size = (float(self._vol_install.capacity) /
-                          1024.0 / 1024.0 / 1024.0)
-        return self._size
-
     def get_dev_type(self):
         if not self._dev_type:
             self._dev_type = _get_dev_type(self._path, self._vol_install, None,
@@ -333,29 +467,10 @@ class _StorageCreator(_StorageBase):
 
     def get_driver_type(self):
         if self._vol_install:
-            if self._vol_install.supports_property("format"):
+            if self._vol_install.supports_format():
                 return self._vol_install.format
         return "raw"
 
-    def validate(self, disk):
-        if disk.device in ["floppy", "cdrom"]:
-            raise ValueError(_("Cannot create storage for %s device.") %
-                             disk.device)
-
-        if self._vol_install:
-            self._vol_install.validate()
-            return
-
-        if self._size is None:
-            raise ValueError(_("size is required for non-existent disk "
-                               "'%s'" % self.get_path()))
-
-        err, msg = self.is_size_conflict()
-        if err:
-            raise ValueError(msg)
-        if msg:
-            logging.warning(msg)
-
     def will_create_storage(self):
         return True
     def get_vol_object(self):
@@ -368,6 +483,28 @@ class _StorageCreator(_StorageBase):
         return False
 
 
+class ManagedStorageCreator(_StorageCreator):
+    """
+    Handles storage creation via libvirt APIs. All the actual creation
+    logic lives in StorageVolume, this is mostly about pulling out bits
+    from that class and mapping them to DeviceDisk elements
+    """
+    def __init__(self, conn, vol_install):
+        _StorageCreator.__init__(self, conn)
+
+        self._pool = vol_install.pool
+        self._vol_install = vol_install
+
+    def create(self, progresscb):
+        return self._vol_install.install(meter=progresscb)
+    def is_size_conflict(self):
+        return self._vol_install.is_size_conflict()
+    def validate(self):
+        return self._vol_install.validate()
+    def get_size(self):
+        return float(self._vol_install.capacity) / 1024.0 / 1024.0 / 1024.0
+
+
 class CloneStorageCreator(_StorageCreator):
     """
     Handles manually copying local files for Cloner
@@ -384,13 +521,19 @@ class CloneStorageCreator(_StorageCreator):
         self._size = size
         self._sparse = sparse
 
+    def get_size(self):
+        return self._size
+
     def is_size_conflict(self):
         ret = False
         msg = None
-        vfs = os.statvfs(os.path.dirname(self._path))
-        avail = vfs[statvfs.F_FRSIZE] * vfs[statvfs.F_BAVAIL]
-        need = long(self._size) * long(1024) * long(1024) * long(1024)
-        if need > avail:
+        if self.get_dev_type() == "block":
+            avail = _get_size(self._path)  # pragma: no cover
+        else:
+            vfs = os.statvfs(os.path.dirname(os.path.abspath(self._path)))
+            avail = vfs.f_frsize * vfs.f_bavail
+        need = int(self._size) * 1024 * 1024 * 1024
+        if need > avail:  # pragma: no cover
             if self._sparse:
                 msg = _("The filesystem will not have enough free space"
                         " to fully allocate the sparse file when the guest"
@@ -401,30 +544,42 @@ class CloneStorageCreator(_StorageCreator):
 
 
             if msg:
-                msg += (_(" %d M requested > %d M available") %
-                        ((need / (1024 * 1024)), (avail / (1024 * 1024))))
+                msg += " "
+                msg += (_("%(mem1)s M requested > %(mem2)s M available") %
+                        {"mem1": (need // (1024 * 1024)),
+                         "mem2": (avail // (1024 * 1024))})
         return (ret, msg)
 
+    def validate(self):
+        if self._size is None:  # pragma: no cover
+            raise ValueError(_("size is required for non-existent disk "
+                               "'%s'" % self.get_path()))
+
+        err, msg = self.is_size_conflict()
+        if err:
+            raise ValueError(msg)  # pragma: no cover
+        if msg:
+            log.warning(msg)  # pragma: no cover
+
     def create(self, progresscb):
         text = (_("Cloning %(srcfile)s") %
                 {'srcfile': os.path.basename(self._input_path)})
 
-        size_bytes = (long(self.get_size()) *
-                      long(1024) * long(1024) * long(1024))
-        progresscb.start(filename=self._output_path, size=long(size_bytes),
+        size_bytes = int(self.get_size() * 1024 * 1024 * 1024)
+        progresscb.start(filename=self._output_path, size=size_bytes,
                          text=text)
 
         # Plain file clone
         self._clone_local(progresscb, size_bytes)
 
     def _clone_local(self, meter, size_bytes):
-        if self._input_path == "/dev/null":
+        if self._input_path == "/dev/null":  # pragma: no cover
             # Not really sure why this check is here,
             # but keeping for compat
-            logging.debug("Source dev was /dev/null. Skipping")
+            log.debug("Source dev was /dev/null. Skipping")
             return
         if self._input_path == self._output_path:
-            logging.debug("Source and destination are the same. Skipping.")
+            log.debug("Source and destination are the same. Skipping.")
             return
 
         # If a destination file exists and sparse flag is True,
@@ -445,11 +600,11 @@ class CloneStorageCreator(_StorageCreator):
             clone_block_size = 1024 * 1024 * 10
             sparse = False
 
-        logging.debug("Local Cloning %s to %s, sparse=%s, block_size=%s",
+        log.debug("Local Cloning %s to %s, sparse=%s, block_size=%s",
                       self._input_path, self._output_path,
                       sparse, clone_block_size)
 
-        zeros = '\0' * 4096
+        zeros = b'\0' * 4096
 
         src_fd, dst_fd = None, None
         try:
@@ -470,15 +625,20 @@ class CloneStorageCreator(_StorageCreator):
                         os.lseek(dst_fd, s, 1)
                     else:
                         b = os.write(dst_fd, l)
-                        if s != b:
+                        if s != b:  # pragma: no cover
                             meter.end(i)
                             break
                     i += s
                     if i < size_bytes:
                         meter.update(i)
-            except OSError as e:
-                raise RuntimeError(_("Error cloning diskimage %s to %s: %s") %
-                                (self._input_path, self._output_path, str(e)))
+            except OSError as e:  # pragma: no cover
+                log.debug("Error while cloning", exc_info=True)
+                msg = (_("Error cloning diskimage "
+                         "%(inputpath)s to %(outputpath)s: %(error)s") %
+                         {"inputpath": self._input_path,
+                          "outputpath": self._output_path,
+                          "error": str(e)})
+                raise RuntimeError(msg) from None
         finally:
             if src_fd is not None:
                 os.close(src_fd)
@@ -486,22 +646,45 @@ class CloneStorageCreator(_StorageCreator):
                 os.close(dst_fd)
 
 
-class ManagedStorageCreator(_StorageCreator):
+class StorageBackendStub(_StorageBase):
     """
-    Handles storage creation via libvirt APIs. All the actual creation
-    logic lives in StorageVolume, this is mostly about pulling out bits
-    from that class and mapping them to VirtualDisk elements
+    Class representing a storage path for a parsed XML disk, that we
+    don't want to do slow resolving of unless requested
     """
-    def __init__(self, conn, vol_install):
-        _StorageCreator.__init__(self, conn)
+    def __init__(self, conn, path, dev_type, driver_type):
+        _StorageBase.__init__(self, conn)
+        self._path = path
+        self._dev_type = dev_type
+        self._driver_type = driver_type
 
-        self._pool = vol_install.pool
-        self._vol_install = vol_install
 
-    def create(self, progresscb):
-        return self._vol_install.install(meter=progresscb)
+    def get_path(self):
+        return self._path
+    def get_vol_object(self):
+        return None
+    def get_vol_xml(self):
+        return None
+    def get_parent_pool(self):
+        return None
+    def get_size(self):
+        return 0
+    def exists(self):
+        return True
+    def get_dev_type(self):
+        return self._dev_type
+    def get_driver_type(self):
+        return self._driver_type
+
+    def validate(self):
+        return
+    def get_vol_install(self):
+        return None
     def is_size_conflict(self):
-        return self._vol_install.is_size_conflict()
+        return (False, None)
+    def is_stub(self):
+        return True
+    def will_create_storage(self):
+        return False
 
 
 class StorageBackend(_StorageBase):
@@ -520,8 +703,8 @@ class StorageBackend(_StorageBase):
             self._path = None
 
         if self._vol_object and not self._parent_pool:
-            raise RuntimeError(
-                "programming error: parent_pool must be specified")
+            raise xmlutil.DevError(
+                "parent_pool must be specified")
 
         # Cached bits
         self._vol_xml = None
@@ -561,15 +744,15 @@ class StorageBackend(_StorageBase):
             if self._vol_object:
                 ret = self.get_vol_xml().capacity
             elif self._path:
-                ret = _stat_disk(self._path)[1]
+                ret = _get_size(self._path)
             self._size = (float(ret) / 1024.0 / 1024.0 / 1024.0)
         return self._size
 
     def exists(self):
         if self._exists is None:
-            if self._path is None:
+            if self._vol_object:
                 self._exists = True
-            elif self._vol_object:
+            elif self._path is None:
                 self._exists = True
             elif (not self.get_dev_type() == "network" and
                   not self._conn.is_remote() and
@@ -610,8 +793,7 @@ class StorageBackend(_StorageBase):
                 return ret
         return None
 
-    def validate(self, disk):
-        ignore = disk
+    def validate(self):
         return
     def get_vol_install(self):
         return None
@@ -619,7 +801,3 @@ class StorageBackend(_StorageBase):
         return (False, None)
     def will_create_storage(self):
         return False
-    def create(self, progresscb):
-        ignore = progresscb
-        raise RuntimeError("programming error: %s can't create storage" %
-            self.__class__.__name__)
